#!/usr/bin/env python
#Author: Mark Feineigle
#Create Date: 2018/02/17
#Last Modify: 2018/02/17
# Simple RSA
# (2017)_Stallings_-_Cryptography_and_network_security_principles_and_practice.pdf
# page 296
#
# M, the block size must be less than n
import json
#import gcd
import euclide2

def keygen(p, q, e, verbose=False):
  if verbose: print "p={}, q={}, e={}".format(p,q,e)
  n = p*q
  phi = (p-1)*(q-1)
  #d = int(gcd.extendedEuclid(e, phi)[1]) #NOTE this only works sometimes
  d = int(euclide2.inverse(e, phi))
  pu = {"e":e, "n":n} #public key
  pr = {"d":d, "n":n} #private key
  return (pu, pr)

def public_or_private(key):
  return "e" if "e" in key.keys() else "d"

def crypt(msg, key, verbose=False):
  k = public_or_private(key)
  if verbose: print "message  = {}".format(msg)
  cipher = [(int((ord(i))**key[k] % key["n"])) for i in msg]
  if verbose: print "cipher   = {}".format(cipher)
  return json.dumps(cipher)

def decrypt(cipher, key, verbose=False, debug=False):
  cipher = json.loads(cipher)
  k = public_or_private(key)
  if debug: print "decipher = {}".format([(int(i**key[k] % key["n"])) for i in cipher])
  msg = "".join([chr(int(i**key[k] % key["n"])) for i in cipher])
  if verbose: print "message  = {}".format(msg)
  return msg


if __name__ == "__main__":
  p = 73
  q = 151
  e = 11
  pu, pr = keygen(p, q, e, True) #pu=public, pr=private
  msg = "How are you?"
  
  print "public key  = {}".format(pu)
  print "private key = {}".format(pr)
  print
  print "public  -> private"
  c1 = crypt(msg, pr, verbose=True) #authenticity
  c2 = crypt(c1, pu, verbose=True) #confidentiality
  c1 = decrypt(c2, pr, verbose=True, debug=True)
  msg = decrypt(c1, pu, verbose=True, debug=True) 
  #print msg
  print
  print "private -> public"
  c1 = crypt(msg, pu, verbose=True)
  c2 = crypt(c1, pr, verbose=True)
  c1 = decrypt(c2, pu, verbose=True, debug=True)
  msg = decrypt(c1, pr, verbose=True, debug=True)
  #print msg
